/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
* PINEWOOD DERBY RACE MANAGER 3-Lane                                              Version 1.0  6/07/2021 *
* Created by: W. van Leeuwen (Billv923@outlook.com)                                                      *
*                                                                                                        *
* This sketch provides the race management features that work in conjunction with the author's Arduino   *
* UNO microcontroller based pinewood derby 3-lane timer running software package PWD_ETTimer03-5a.ino.   *
* It provides for: (1) importing the user's race roster and generating the race plan, (2) controlling    *
* the timer hardware during the race, (3) saving the race results and creating the race results and      *
* standings reports at completion of the race. Race plans are generated using the "Partial Perfect-N"    *
* (PPN) concept where all racers run once on each lane and against different racers each time as         *
* determined by the number of racers in the group.                                                       *
*                                                                                                        *
* NOTE: This program requires the user to provide a roster file in comma-delimited values (.csv) format  *
*       containing the racer names and car numbers. The program will use the roster file to create a     *
*       race plan using the PPN Table. Refer to the user manual for the required roster data file format.*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
   
  import processing.serial.*;                        //Use the serial library
  import static javax.swing.JOptionPane.*;           //Use the option box dialog library

//Variables
  int PageSelect = 1;                                //1=Setup page, 2=Race page, 3=View/Print page
  int Num_Lanes = 3;                                 //Number of lanes
  int h, i, j, k, n, x, y;                           //Working integer variables
  int L1, L2, L3;                                    //Working Lane car number variables
  int R1, R2, R3;                                    //Working Racer Number variables
  int Rx = 250;                                      //Default Background color Red value
  int Gx = 250;                                      //Default Background color Green value
  int Bx = 150;                                      //Default Background color Blue value
  int Ry = 140;                                      //Racing Screen Background color Red value
  int Gy = 200;                                      //Racing Screen Background color Green value
  int By = 230;                                      //Racing Screen Background color Blue value
  int Rz = 170;                                      //Result Screen Background color Red value
  int Gz = 250;                                      //Result Screen Background color Green value
  int Bz = 150;                                      //Result Screen Background color Blue value
  int Yr, Mo, Day, Hr, Min;                          //Integer variables for creating date-time based filename
  int Race_State = 0;                                //0=Not Ready, 1=Ready, 2=Racing, 3=Finished, 4=Complete
  int HeatNum = 0;                                   //Heat (race) number
  int Racers;                                        //Total number of racers in roster
  int MaxRacers = 40;                                //Maximum number of racers allowed
  int Max_Heats = 40;                                //Maximum number of heats allowed
  int Heat_Number = 0;                               //Heat number selected/displayed
  int [] Place = new int[3];                         //Array for storing lane finish order
  int [] Car = new int[3];                           //Array for storing car assignment to lanes
  int Total_Heats;                                   //Total number of heats in the race plan
  int Index1 = -1;                                   //Variable - Table row index pointer
  int Index2 = -1;                                   //Variable - Table row index pointer
  int Itime1, Itime2;                                //Integer time variables used to calculate total time
  int [] kar = new int[40];                          //Temp array used to check repeat car numbers
  int lf = 10;                                       //ASCII Line feed
  boolean RosterFile_Selected = false;               //Denotes when user has selected a Roster file to use
  boolean RosterLoadedflag = false;                  //Denotes when user has loaded a roster file
  boolean Errorflag1 = false;                        //Denotes when an error  has been detected
  boolean Errorflag2 = false;                        //Denotes when an error  has been detected
  boolean PlanCreatedflag = false;                   //Denotes when user has created a race plan
  boolean ComPort_Avail = false;                     //Denotes when com port is available. Default to false
  boolean Rerunflag = false;                         //Used to indicate when Rerun Button is to be active
  boolean NextHeatflag = true;                       //Used to indicate when Next Heat button is to be active
  boolean RaceCompflag = false;                      //Used to indicate when all heats have been completed
  boolean [] Mask = new boolean[4];                  //Array for storing lane mask status
  boolean [] Heat_Comp = new boolean[40];            //Array for storing completion state (true/false) of each heat
  boolean Gaterelflag = false;                       //Gate release flag
  boolean RaceResultsAvailflag = false;              //Denotes when race results are available
  boolean StandingsAvailflag = false;                //Denotes when race standings are available
  boolean GrpStdngsAvailflag = false;                //Denotes when group standings are available
  boolean Sortflag;                                  //Used in standings table sort routines
  float [] Time = new float[3];                      //Array for storing current lane time values
  float Time1, Time2, Time3;                         //Variables for temp storage of time values
  String RP_Text1 = "|Lane| Name               | Car# |";
  String RP_Text2 = "----------------------------------";
  String Spaces = "                              ";  //String of 30 spaces
  String RP_Title = "User Selectable Title Goes Here"; //Default Racing Page title
  String RP_Header = "";                             //String variable for Race Plan header text
  String Roster_Filename = "";                       //String variable for the User's Roster data file name
  String Filename1;                                  //String variable for the Race Plan file name in rtf format
  String Filename2;                                  //String variable for the Race Results file name in csv format
  String Filename3;                                  //String variable for the Overall Race Standing file name in csv format
  String Filename4;                                  //String variable for the Group Race Standing file name in csv format
  String File_Path;                                  //String variable for the file path
  String COMx, COMlist = "";                         //COM Port check/select routine variables
  String Portname = "";                              //COM Port name variable
  String Str1 = "";                                  //Working string variable
  String Str2 = "";                                  //Working string variable
  String Str3 = "";                                  //Working string variable
  String Name;                                       //Working string variable
  String [] Driver = new String[3];                  //Working string variable
  String N1, N2, N3;                                 //Lane Name assignment variables
  String prompt = "";                                //Variable used for storing popup message text
  String inString = "";                              //Input string from serial port
  String Time_String = "";                           //String variable for temp storage of time values
  String Date_String;                                //String variable of date/time value used in file name
  String Msg1 = "";                                  //String variable to hold lane sensor error message
  String Msg1a = "Lane sensor(s)";
  String Msg1b = " not ready. Fix problem & click Timer Reset";
  String Msg2 = "Gate (Start) switch not ready, Fix problem & click Timer Reset";
  String Msg3 = "Click Next Heat or Rerun Heat to prep for next race";
  String Msg4 = "Click Next Heat button to prep for next race";
  String Msg5 = "Click Timer Reset to run a practice heat or Next Heat button to prep for first official heat";
  String Msg6 = "Remote RESET detected";
  String Msg7 = "RACE COMPLETE - Select Results page to view/print results - Restart program to run a new race";
  String Msg12 = "NOTICE: A maximum of two lanes can be masked at a time.\nIf you want to mask this lane, you must unmask an other lane first.";
  String Msg13 = "Race setup has not yet been completed.\nGo to Setup Page and complete setup.";
  PFont font;                                        //Set font
  Table PPN_Table;                                   //Define table named 'PPN_Table'
  Table Roster;                                      //Define table named 'Roster'
  Table Raceplan;                                    //Define table named 'Raceplan'
  Table Results;                                     //Define table named 'Results'
  Table Standings;                                   //Define table named 'Standings'
  PImage logo;                                       //Define an image variable as 'logo'
  Serial myPort;                                     //Set the Serial port name to myPort
  PrintWriter RP_output;                             //Create an output variable for writing to the Race Plan file
  
//*********************************************************************************************************
void setup() {
  size(1200, 700);                                   //Set display area in pixels
  logo = loadImage("BillV_Logo.png");                //Load "Bill V's Derby Timers" Logo
  
  //Create unique date-time based file name so that a new file is created each time the program is run.
  //This will prevent previous files from being accidentally overwritten. Note that by default the file
  //is saved to the same folder/directory that contains this sketch.
  Yr = year();                                       //Get year value from PC clock
  Mo = month();                                      //Get month value from PC clock
  Day = day();                                       //Get day value from PC clock
  Hr = hour();                                       //Get hour value from PC clock
  Min = minute();                                    //Get minute value from PC clock
  Date_String = nf(Yr,4)+nf(Mo,2)+nf(Day,2)+nf(Hr,2)+nf(Min,2);
  File_Path = "C:/Users/";                           //Default File path for saving race documents
  Filename1 = File_Path + "Race_Plan_"+Date_String+".doc";
  Filename2 = File_Path + "Race_Results_"+Date_String+".csv";
  Filename3 = File_Path + "Race_Standings_Overall_"+Date_String+".csv";
  Filename4 = File_Path + "Race_Standings_Group_"+Date_String+".csv";
  
  //Create the Raceplan table (3 lanes).
  Raceplan = new Table();                            //Name the new table
  Raceplan.addColumn("Heat", Table.INT);             //Add a column for Heat number
  Raceplan.addColumn("L1Car", Table.STRING);         //Add a column for Lane1 Car Number
  Raceplan.addColumn("L1Name", Table.STRING);        //Add a column for Lane1 Name
  Raceplan.addColumn("L2Car", Table.STRING);         //Add a column for Lane2 Car Number
  Raceplan.addColumn("L2Name", Table.STRING);        //Add a column for Lane2 Name
  Raceplan.addColumn("L3Car", Table.STRING);         //Add a column for Lane3 Car Number
  Raceplan.addColumn("L3Name", Table.STRING);        //Add a column for Lane3 Name
  
  //Create the Results table. The table is used to store data from each heat
  Results = new Table();                             //Name the new table
  Results.addColumn("Heat", Table.INT);              //Add a column for Heat number
  Results.addColumn("L1Car", Table.STRING);          //Add a column for Lane1 Car Number
  Results.addColumn("L1Time", Table.FLOAT);          //Add a column for Lane1 time
  Results.addColumn("L2Car", Table.STRING);          //Add a column for Lane2 Car Number
  Results.addColumn("L2Time", Table.FLOAT);          //Add a column for Lane2 time
  Results.addColumn("L3Car", Table.STRING);          //Add a column for Lane3 Car Number
  Results.addColumn("L3Time", Table.FLOAT);          //Add a column for Lane3 time
  
  //Create Standings stable. Used to record race standings of each car at race completion
  Standings = new Table();                           //Name the new table
  Standings.addColumn("Position", Table.INT);        //Add a column named Position (Position standing based on Time)
  Standings.addColumn("CarNo", Table.STRING);        //Add a column named CarNo
  Standings.addColumn("Driver", Table.STRING);       //Add a column named Driver (Driver name)
  Standings.addColumn("Group", Table.STRING);        //Add a column named Group (Group name)
  Standings.addColumn("TotalTime", Table.FLOAT);     //Add a column named TotalTime (Sum of all race times)
  Standings.addColumn("HeatsRan", Table.INT);        //Add a column named HeatsRan (Number of Heats ran)
  Standings.addColumn("AvgTime", Table.FLOAT);       //Add a column named AvTime (Average Time = TotalTime/HeatsRan)
  Standings.addColumn("FastTime", Table.FLOAT);      //Add a column named FastTime (Fastest time ran)
  Standings.addColumn("IntTime", Table.INT);         //Add a column named IntTime (Total time in Integer format)
  
  //Load the PPN table
  PPN_Table = loadTable("PPN_Table-3.csv", "header"); //Load 3-lane PPN table from file

  //Set the font. The font must be located in the sketch's "data" directory to load successfully
  font = loadFont("ArialNarrow-Bold-48.vlw");
  textFont(font, 60);
  PageSelect = 1;                                    //Set selection to Setup page
  Init_Setup_Page();                                 //Call subroutine to initialize the setup page
  
} //End of "Setup" routine

//*********************************************************************************************************
void draw() {  //Note: Things aren't displayed on the screen until the draw loop completes.
  if (ComPort_Avail && PageSelect==2 ) {
    if (inString.indexOf("NRD") >= 0) {              //Check for receipt of 'Not Ready' state message
      Race_State = 0;                                //Set the race state to 0 (0=not ready)
      Set_Race_State();                              //and call subroutine to display the race state
      if (Heat_Number > 0) {                         //Don't allow if no heats have been ran yet
        NextHeatflag = false;                        //Disable Next Heat button
        Draw_NextHeat_Button();                      //Call subroutine to update Next Heat button display
      }
      Rerunflag = false;                             //Set Rerunflag to disable Rerun Heat button
      Draw_RerunHeat_Button();                       //Call subroutine to update Rerun Heat button display
      inString = "";                                 //Clear 'Not Ready' message once processed
    }
    if (inString.indexOf("RDY") >= 0) {              //Check for receipt of 'Ready' state from timer
      Race_State = 1;                                //Set the race state to 1 (1=ready)
      Set_Race_State();                              //Call subroutine to display the race state
      Clear_Data();                                  //Call subroutine to clear data from text boxes
      Clear_Msgbox();                                //Call subroutine to initialize/clear Message box
      if (Heat_Number > 0) {                         //Don't allow if no heats have been ran yet
        NextHeatflag = false;                        //Disable Next Heat button
        Draw_NextHeat_Button();                      //Call subroutine to update Next Heat button display
      }
      Rerunflag = false;                             //Toggle the Rerunflag
      Draw_RerunHeat_Button();                       //Call subroutine to update Rerun Heat button
      inString = "";                                 //Clear 'Ready' message once processed
    }
    if (inString.indexOf("RAC") >= 0) {              //Check for receipt of 'Racing' state
      Race_State = 2;                                //Set the race state to 2 (2=racing)
      Set_Race_State();                              //and call subroutine to display the race state
      inString = "";                                 //Clear 'Racing' message once processed
    }
    if (inString.indexOf("FIN") >= 0) {              //Check for receipt of 'Finished' state
      Race_State = 3;                                //Set the race state to 3 (3=finished)
      Set_Race_State();                              //and call subroutine to display the race state
      if (Heat_Number == 0) {                        //Check if practice heat
        Disp_Message(5);                             //Call subroutine to display initial start up message
        Rerunflag = false;                           //Toggle the Rerunflag to FALSE
      }
      else {
        Disp_Message(3);                             //Call subroutine to display Next/Rerun Heat message
        Rerunflag = true;                            //Toggle the Rerunflag to TRUE
      }
      Draw_RerunHeat_Button();                       //Call subroutine to update Rerun Heat button
      NextHeatflag = true;                           //Enable Next Heat button
      Draw_NextHeat_Button();                        //Call subroutine to update Next Heat button display
      inString = "";                                 //Clear 'Finished' message once processed
    }
    if (inString.indexOf("GSW") >= 0) {              //Check for receipt of Gate Switch error code
      Disp_Message(2);                               //Call subroutine to display Gate Switch error message
      inString = "";                                 //Clear error message once processed
    }
    if (inString.indexOf("TRK") >= 0) {              //Check for receipt of Track Sensor error code
      i = inString.length() - 2;                     //Get length of inString minus cr/lf characters
      Str1 = inString.substring(4, i);               //Extract lane sensor numbers from inString
      Msg1 = Msg1a + Str1 + Msg1b;                   //Construct the error message to display
      Disp_Message(1);                               //Call subroutine to display track sensor error message
      inString = "";                                 //Clear error message once processed
    }
    if (inString.indexOf("RST") >= 0) {              //Check for receipt of Remote RESET message
      for (i=0; i<Num_Lanes; i++) {                  //Loop through all lanes
        Mask[i] = false;                             //and set its Mask flag to false
      }
      Set_Mask_State();                              //Refresh display with lane mask states
      Disp_Message(6);                               //Call subroutine to display Remote Reset message
      inString = "";                                 //Clear message once processed
    }
    if (inString.indexOf("Time") >= 0) {             //Check for receipt of Race times
      Time_String = inString;                        //and save it to a temp holding variable
      //Note: This string has the following format: "Time: a.aaaa b.bbbb c.cccc"
      //      where a.aaaa through c.cccc are the lanes times for lanes 1 - 3 respectively
      Str1 = Time_String.substring(7, 13);           //Extract race time string for lane 1
      Time[0] = float (Str1);                        //and convert it to a floating point number
      Str1 = Time_String.substring(14, 20);          //Extract race time string for lane 2
      Time[1] = float (Str1);                        //and convert it to a floating point number
      Str1 = Time_String.substring(21, 27);          //Extract race time string for lane 3
      Time[2] = float (Str1);                        //and convert it to a floating point number
      inString = "";                                 //Clear Race times string once processed
    
      //Determine the lane finish order (place) based on finish times
      for (i=0; i<Num_Lanes; i++) {                  //First set all Place holders to last (3'd) place
        Place[i] = Num_Lanes;                        //(Last place = number of lanes)
      }
      for (i=0; i<Num_Lanes; i++) {                  //Now compare each time to the other times and
        for (j=0; j<Num_Lanes; j++) {                //decrement its place value each time it is less
          if (Time[i] < Time[j]) {                   //A value of 9.9999, if present, gets set to last place thus
            Place[i]--;                              //maintaining correct finish order for the valid lanes
          }
        }
      }
      //Write the time values into their respective text boxes on the screen
      Clear_Data();                                  //Call subroutine to clear any data from text boxes
      textSize(55);                                  //Set text size to 55 pixels
      textAlign(CENTER);                             //Set text alignment to center
      fill(255);                                     //Set text color to white
      for (i=0; i<Num_Lanes; i++) {                  //Loop through all lanes
        if (!Mask[i]) {                              //If lane mask is NOT set
          Str1=nf(Time[i],1,4);                      //Convert so it won't round to 3 decimal places
          if (Time[i] != 9.9999) text(Str1,250+i*350,290);   //Check for a valid time value and write if so
          else text("- - - - -",255+i*350,285);      //Or write dashes into the lane 1 text box
        }
      }
      //Write the finish order values into their respective text boxes on the screen if valid
      textSize(130);                                 //Set text size to 130 pixels
      for (i=0; i<Num_Lanes; i++) {                  //Loop through all lanes
        if (Time[i] != 9.9999) text(Place[i],250+i*350,225); //Write Place value if it is valid
      }
    
      //Write times to the Results table
      if (Heat_Number > 0) {                         //Do only if a valid heat number
        Index1 = Heat_Number-1;
        Results.setFloat(Index1, "L1Time", Time[0]); //Save Lane 1 time to the Results table
        Results.setFloat(Index1, "L2Time", Time[1]); //Save Lane 2 time to the Results table
        Results.setFloat(Index1, "L3Time", Time[2]); //Save Lane 3 time to the Results table
        Heat_Comp[Index1] = true;                    //Set 'Heat Complete' flag for this heat to TRUE
        //Save (re-save) all results to the datafile each time new results are available. Note that by default
        //the file is saved to the same folder/directory that contains this sketch.
        saveTable(Results, Filename2);               //Save/resave race results
        RaceResultsAvailflag = true;                 //Set flag to signify race results are available
      }
      delay(100);                                    //Wait awhile before we loop again
    }                                                //End of Receipt of race times routine
    
    //Test for race completion of all heats
    if (Race_State == 3) {                           //Do only if in "Finished" race state
      RaceCompflag = true;                           //Preset Completion flag to TRUE
      for (i=0; i<Total_Heats; i++) {                //Loop through all heats
        if (!Heat_Comp[i]) RaceCompflag = false;     //And set flag to FALSE if any heat is not complete
      }
      if (RaceCompflag) {                            //Test for completion of race
        Populate_Standings_Table();                  //Complete the Standings table
        Disp_Message(7);                             //Call subroutine to display Race Complete message
        NextHeatflag = false;                        //Disable Next heat button
        Draw_NextHeat_Button();                      //And update display
        RaceCompflag = false;                        //Clear flag so we don't continuously rerun this routine
        Race_State = 4;                              //Set race state to "Complete"
      }
    }
  }
  
} //End of "draw" loop

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*                                                SUBROUTINES                                               *
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void serialEvent(Serial p) {                         //Serial data handling routine
  inString = p.readString();
}

void Init_Setup_Page() {                             //Initialize the Setup page
  background(Rx, Gx, Bx);                            //Clear screen with Setup screen background color
  Draw_Tabs();                                       //Call subroutine to draw tabs
  textAlign(CENTER);                                 //Set text alignment to center
  textSize(40);                                      //Set text size to 40 pixels
  fill(255, 0, 170);                                 //Set Title color
  text("SETUP",600,75);                              //Display Title
  image(logo, 1080, 640);                            //And display it in lower right corner
  textAlign(CENTER);                                 //Set text alignment to center
  textSize(25);                                      //Set text size to 25 pixels
  fill(0);                                           //Set color to black
  text("Use this screen to:",600,110);
  text("(1) Select timer COM port, (2) import the race roster, (3) create the race plan, (4) update race page title.",600,140);
  text("NOTE: The first 3 steps above must be performed in order to start the race.",600,170);
  Draw_SelectCom_Button();                           //Call subroutine to display Check Comms button
  Draw_LoadRoster_Button();                          //Call subroutine to display Load Roster button
  Draw_CreatePlan_Button();                          //Call subroutine to display Create Plan button
  Draw_UpdateTitle_Button();                         //Call subroutine to display Update Header button
  Draw_Reset_Button();                               //Call subroutine to display RESET button
}

void Init_Race_Page() {                              //Initialize the Racing page
  background(Ry, Gy, By);                            //Clear screen with Racing screen background color
  Draw_Tabs();                                       //Call subroutine to draw tabs
  textAlign(CENTER);                                 //Set text alignment to center
  textSize(40);                                      //Set text size to 40 pixels
  fill(255, 0, 170);                                 //Set Title color
  text(RP_Title,600,65);                             //Display Racing Page header - can be updated by end user
  image(logo, 1080, 640);                            //And display it in lower right corner
  fill(0);                                           //Set color to black
  text("Lane 1",250,110);                            //Lane 1 legend
  text("Lane 2",600,110);                            //Lane 2 legend
  text("Lane 3",950,110);                            //Lane 3 legend
  textSize(30);                                      //Set text size to 30 pixels
  fill(245, 255, 0);                                 //Set text color to yellow
  textAlign(LEFT);                                   //Set text alignment to left
  text("Mask",280,337);                              //Lane 1 Mask legend
  text("Mask",630,337);                              //Lane 2 Mask legend
  text("Mask",980,337);                              //Lane 3 Mask legend
  fill(Ry, Gy, By);                                  //Set color same as background
  rect(40,420,260,190,10);                           //Draw rectangle for Race Status box
  rect(320,420,840,190,10);                          //Draw rectangle for Race Control box
  rect(320,420,100,190,10);                          //Draw rectangle for Reset Timer box
  fill(0);                                           //Set color back to black
  text("Race Status",100,450);                       //Draw Race Status box legend
  text("Race Control",550,450);                      //Draw Race Control box legend
  text("Messages:",40,640);                          //Draw message box legend
  fill(255, 0, 0);                                   //Set color to Red
  ellipse(370,530,50,50);                            //Draw Reset Button
  textAlign(CENTER);                                 //Set text alignment to center
  textSize(25);                                      //Set text size to 25 pixels
  text("Timer",370,500);                             //Draw Reset Button legend
  text("Reset",370,580);                             //Draw Reset Button legend
  fill(0,0,255);                                     //Set color to Blue
  ellipse(890,530,50,50);                            //Draw Terminate Race button
  text("Terminate",890,500);                         //Draw Terminate Race button legend
  text("Race",890,580);                              //Draw Terminate Race button legend
  fill(0);                                           //Set color back to black
  text("Heat #",1040,485);                           //Draw Heat Box text header
  fill(100);                                         //Set rectangle color
  rect(1110,490,30,80);                              //and draw Up/Down increment box
  fill(0,0,255);                                     //Set color to Blue
  triangle(1110, 525, 1140, 525, 1125, 495);         //Draw Up Arrow
  triangle(1110, 535, 1140, 535, 1125, 565);         //Draw Down Arrow
  //Initialize text & message boxes by calling their respective subroutines
  Clear_Data();                                      //Call subroutine to initialize lane time text boxes
  Clear_Msgbox();                                    //Call subroutine to initialize Message box
  Clear_Namebox();                                   //Call subroutine to initialize names & car number boxes
  Draw_Heatbox();                                    //Call subroutine to initialize Heat Count box
  Set_Mask_State();                                  //Call subroutine to set lane masking states
  Set_Race_State();                                  //Call subroutine to initialize Race Status box
  Draw_NextHeat_Button();                            //Call subroutine to draw the Next Heat button
  Draw_RerunHeat_Button();                           //Call subroutine to draw the Rerun Heat button
  Draw_GateRel_Button();                             //Call subroutine to draw the Gate Release button
  if (Heat_Number>0 && PlanCreatedflag) {
    Display_CarDriver();                             //Re-display racer info
    Redisplay_RaceResults();                         //Re-display race results if any
  }
  else Disp_Message(5);                              //Call subroutine to display initial race start message
}

void Init_Results_Page() {                           //Initialize the Results page
  background(Rz, Gz, Bz);                            //Clear screen with Results page background color
  Draw_Tabs();                                       //Call subroutine to draw tabs
  textAlign(CENTER);                                 //Set text alignment to center
  textSize(40);                                      //Set text size to 40 pixels
  fill(255, 0, 170);                                 //Set Title color
  text("RACE RESULTS",600,75);                       //Display Title
  image(logo, 1080, 640);                            //And display it in lower right corner
  textAlign(CENTER);                                 //Set text alignment to center
  textSize(25);                                      //Set text size to 25 pixels
  fill(0);                                           //Set color to black
  text("Use this screen to view & print the race documents.",600,120);
  Draw_ViewRoster_Button();                          //Call subroutine to draw the View Roster button
  Draw_ViewPlan_Button();                            //Call subroutine to draw the View Race Plan button
  Draw_ViewRace_Results_Button();                    //Call subroutine to draw the View Race Results button
  Draw_View_Ovl_Standings_Button();                  //Call subroutine to draw the View Overall Standings button
  Draw_View_Grp_Standings_Button();                  //Call subroutine to draw the View Group Standings button
}

void Draw_Tabs() {
  noStroke();                                        //Turn off shape border line
  fill(200);                                         //Set color to gray
  rect(0,0,1200,30);                                 //Draw top border
  fill(Rx, Gx, Bx);                                  //Set color to Setup Screen
  rect(20,5,100,25,10,10,0,0);                       //Draw Setup Screen tab
  fill(Ry, Gy, By);                                  //Set color to Racing Screen
  rect(120,5,100,25,10,10,0,0);                      //Draw Racing Screen tab
  fill(Rz, Gz, Bz);                                  //Set color to Results Screen
  rect(220,5,100,25,10,10,0,0);                      //Draw Results Screen tab
  textAlign(CENTER);                                 //Set text alignment to center
  textSize(15);                                      //Set text size to 15 pixels
  fill(0);                                           //Set Text color to black
  text("Setup",70,23);                               //Write text for Setup screen tab
  text("Racing",170,23);                             //Write text for Racing screen tab
  text("Results",270,23);                            //Write text for Results screen tab
  stroke(0);                                         //Turn on shape border line
}

void mouseClicked() {                                //Mouse Click handling routine
  //This routine is called when the mouse button is clicked. Mouse X & Y coordinates are tested to see
  //if the mouse pointer is over any of the clickable buttons
  x = mouseX;                                        //Get mouse X-coordinate
  y = mouseY;                                        //Get mouse Y-coordinate
  //Test for Setup Tab Select
  if ((x>20) && (x<120) && (y>5) && (y<30)) {        //Setup tab clicked?
    PageSelect = 1;                                  //Set screen selection to Setup page
    Init_Setup_Page();                               //Call subroutine to initialize the setup page
  }
  if ((x>120) && (x<220) && (y>5) && (y<30)) {       //Racing tab clicked?
    PageSelect = 2;                                  //Set screen selection to Racing page
    Init_Race_Page();                                //Call subroutine to initialize the racing page
  }
  if ((x>220) && (x<320) && (y>5) && (y<30)) {       //Racing tab clicked?
    PageSelect = 3;                                  //Set screen selection to Results page
    Init_Results_Page();                             //Call subroutine to initialize the Results page
  }
  if (PageSelect==1) {                               //Do if Setup page is selected
    //Test for Select Comm button click
    if ((x>50) && (x<130) && (y>180) && (y<240)) {   //Select Comm button clicked?
      if (!ComPort_Avail) Select_Comm();             //Call subroutine to select com port
      if (ComPort_Avail) {                           //Do if COM port available
        myPort.write("C");                           //Send the letter C for comm check request
        myPort.write(lf);                            //followed by a line-feed
        fill(Rx, Gx, Bx);                            //Set color to setup screen background
        noStroke();                                  //Disable object draw outline
        rect(140,180,860,60);                        //Clear any previous com check messages
        stroke(0);                                   //Enable object draw outline
        i = millis() + 2000;                         //Get CPU time + 2 seconds
        while (millis() < i) {                       //Loop for max of 2 seconds
          if (inString.indexOf("@") >= 0) {          //Did timer send acknowledge?
            textSize(20);                            //Set text size to 20 pixels
            textAlign(LEFT);                         //Set text alignment to center
            fill(0);                                 //Set text color to black
            text("Comm Check OK on Port: " + Portname, 150, 215);
            break;                                   //exit the while loop
          }
        }
      }
    }
    //Test for Load Roster button click 
    if ((x>50) && (x<130) && (y>280) && (y<340)) {   //Load Roster button clicked?
      if (!RosterLoadedflag) {                       //Check if roster is loaded
        Load_Roster();                               //Call load roster subroutine
        RosterLoadedflag = true;                     //Set flag to indicate roster is loaded
        Draw_LoadRoster_Button();                    //Call subroutine to redraw load roster button
        Draw_CreatePlan_Button();                    //Call subroutine to redraw Create Plan button
        Draw_Reset_Button();                         //Call subroutine to redraw Reset button
      }
      else {
        prompt = "Roster already loaded.\nPlease select RESET if you want to start over.";
        showMessageDialog(frame,prompt);             //Display the prompt for acknowledgement
      }
    }
    //Test for Create Plan button click
    if ((x>50) && (x<130) && (y>380) && (y<440)) {   //Create Plan button clicked?
      if (RosterLoadedflag && !PlanCreatedflag) {    //Check if roster is loaded & race plan created
        Create_RacePlan();                           //Call Create Race Plan subroutine
        PlanCreatedflag = true;                      //Set flag to indicate race plan has been created
        Draw_CreatePlan_Button();                    //Call subroutine to redraw CreatePlan button
      }
      else if (!RosterLoadedflag) {                  //Do if roster not loaded
        prompt = "Roster has not been loaded.\nPlease load a roster first";
        showMessageDialog(frame,prompt);             //Display the prompt for acknowledgement
      }
      else if (PlanCreatedflag) {
        prompt = "Looks like a roster has already been created.\nPlease select RESET if you want to start over.";
        showMessageDialog(frame,prompt);             //Display the prompt for acknowledgement
      }
    }
    //Test for Update Title button click
    if ((x>50) && (x<130) && (y>480) && (y<540)) {   //Update Title button clicked?
      Str1 = showInputDialog("Enter the Race Page title below");
      if (Str1!=null || !COMx.isEmpty()) RP_Title = Str1;
      Draw_UpdateTitle_Button();                     //Call subroutine to redraw update title button
    }
    //Test for Setup Page RESET button click
    if ((x>50) && (x<130) && (y>580) && (y<640)) {   //Print Plan button clicked?
      Roster.clearRows();                            //Remove all rows from the Roster table
      Raceplan.clearRows();                          //Remove all rows from the Race Plan table
      RosterLoadedflag = false;                      //Set flag to indicate roster has been cleared
      Draw_LoadRoster_Button();                      //Call subroutine to redraw load roster button
      PlanCreatedflag = false;                       //Set flag to indicate race plan has been cleared
      Draw_CreatePlan_Button();                      //Call subroutine to redraw Create Plan button
      Draw_UpdateTitle_Button();                     //Call subroutine to redraw Update header button
      textSize(20);                                  //Set text size to 20 pixels
      textAlign(LEFT);                               //Set text alignment to center
      fill(Rx, Gx, Bx);                              //Set text color background
      noStroke();                                    //Disable drawing of rectangle outline
      rect(145,280,1050,160);                        //Erase displayed text messages
      stroke(0);                                     //Enable drawing of rectangle outline
      RosterFile_Selected = false;                   //Reset Roster Selected flag
      prompt = "Ready to load new roster";
      showMessageDialog(frame,prompt);               //Display the prompt for acknowledgement
    }
  }
  if (PageSelect==2) {                               //Do if Racing page is selected
    //Test for racing page RESET button and send the reset command to the Arduino if so.
    if ((x>345) && (x<395) && (y>505) && (y<555)) {
      if (ComPort_Avail) {                           //Do only if COM port available
        myPort.write("R");                           //Send the letter R
        myPort.write(lf);                            //followed by a line-feed
        NextHeatflag = true;                         //Enable Next heat button (in case it was disabled)
        Draw_NextHeat_Button();                      //And update display
      }
      else {                                         //Do if COM port not available
        prompt = "COM Port has not yet been established.\nGo to Setup Page and select timer COM port.";
        showMessageDialog(frame,prompt);             //Display Dialog
      }
    }
    //Test for Next Heat select button and perform necessary actions
    if ((x>445) && (x<525) && (y>500) && (y<560)) {
      if (ComPort_Avail && PlanCreatedflag) {        //Check if setup complete to run a race
        if (NextHeatflag) {                          //Only allow if Next heat flag is TRUE
          for (i=0; i<Max_Heats; i++) {              //Loop through all heats
            if (!Heat_Comp[i]) {                     //Check for the first un-ran heat
              Heat_Number = i + 1;                   //And set the heat number
              break;                                 //Exit of the 'for' loop
            }
          }
          if (Heat_Number > Max_Heats) Heat_Number--;  //Limit Heat Number to max heats
          Display_CarDriver();                       //Display car & driver info for the current heat
          Draw_Heatbox();                            //Redisplay new heat value
          myPort.write("R");                         //Send the Reset command to Arduino to ready the timer
          myPort.write(lf);                          //followed by a line-feed
        }
      }
      else {
        showMessageDialog(frame,Msg13);              //Display error message dialog
      }
    }
    //Test for Rerun Heat select button and toggle Rerunflag
    if ((x>565) && (x<645) && (y>500) && (y<560)) {
      if (Rerunflag) {                               //Only allow if Rerun flag is TRUE
        Str1 = "RERUN SELECT";                       //Set Dialog header
        prompt = "RERUN CURRENTLY SELECTED HEAT\nAre you sure?";   //Build the prompt string
        int input = showConfirmDialog(null, prompt, Str1, YES_NO_OPTION);  //Prompt user for heat rerun: 0=yes, 1=no
        if (input == 0) {                            //Check if YES is selected and do following
          myPort.write("R");                         //Send the Reset command to Arduino to ready the timer
          myPort.write(lf);                          //followed by a line-feed
        }
      }
    }
    //Test for Gate Release button (Not implemented in this version)
    if ((x>685) && (x<765) && (y>500) && (y<560)) {
      prompt = "Gate Release feature not\nimplemented in this version.";
      showMessageDialog(frame,prompt);               //Display Dialog
    }
    //Test for Terminate Race button and set Terminate flag if so.
    if ((x>865) && (x<915) && (y>505) && (y<555)) {
      Str1 = "TERMINATE RACE SELECT";                //Set Dialog header
      if (Race_State<4 && PlanCreatedflag) {         //Do only if race is not already complete
        prompt = "TERMINATE RACE\nAre you sure?\nAll heats have not yet completed."; //Build the prompt string
        int input = showConfirmDialog(null, prompt, Str1, YES_NO_OPTION);  //Prompt user for race termination: 0=yes, 1=no
        if (input == 0) {                            //Check if YES is selected and do following
          Populate_Standings_Table();                //Complete the Standings table
          prompt = "RACE TERMINATED\nResults have been saved\nRestart program to run a new race";
          showMessageDialog(frame,prompt);           //Display Dialog
          exit();
        }
      }
      else {
        prompt = "EXIT PROGRAM\nAre you sure?";      //Build the prompt string
        int input = showConfirmDialog(null, prompt, Str1, YES_NO_OPTION);  //Prompt user for race exit: 0=yes, 1=no
        if (input == 0) {                            //Check if YES is selected and do following
          prompt = "Restart program to run a new race";
          showMessageDialog(frame,prompt);           //Display Dialog
          exit();
        }
      }
    }
    //Test for Up Arrow button and increment Heat Number value by one if so.
    if ((x>1110) && (x<1140) && (y>495) && (y<525)) {
      if (ComPort_Avail && PlanCreatedflag) {        //Check if setup complete to run a race
        Heat_Number++;                               //Increment Heat Number
        if (Heat_Number > Total_Heats) Heat_Number--;//Limit Heat Number to total heats
        Display_CarDriver();                         //Display car & driver info for the current heat
        Redisplay_RaceResults();                     //Display time & finish order info
        Draw_Heatbox();                              //and redisplay new heat value
      }
      else {
        showMessageDialog(frame,Msg13);              //Display error message dialog
      }
    }
    //Test for Down Arrow button and decrement Heat Number value by one if so.
    if ((x>1110) && (x<1140) && (y>535) && (y<565)) {
      if (ComPort_Avail && PlanCreatedflag) {        //Check if setup complete to run a race
        Heat_Number--;                               //Decrement Heat Number
        if (Heat_Number < 1) Heat_Number = 1;        //but don't let it get less than 1
        Display_CarDriver();                         //Display car & driver info for the current heat
        Redisplay_RaceResults();                     //Display time & finish order info
        Draw_Heatbox();                              //and redisplay new heat value
      }
      else {
        showMessageDialog(frame,Msg13);              //Display error message dialog
      }
    }
    //Test for lane 1 mask checkbox and set/reset lane 1 mask flag
    if ((x>230) && (x<270) && (y>305) && (y<345)) {
      if (ComPort_Avail && PlanCreatedflag) {        //Check if setup complete to run a race
        Mask[0] = !Mask[0];                          //Toggle lane 1 mask state
        if (Mask[0] && Mask[1] && Mask[2]) {         //Are more than 2 lanes masked?
          showMessageDialog(frame,Msg12);            //Display Dialog
          Mask[0] = false;                           //Clear Lane 1 mask flag
        }
        Set_Mask_State();                            //and update the display
        Send_Mask_State();                           //Send mask status to Arduino
      }
      else {
        showMessageDialog(frame,Msg13);              //Display error message dialog
      }
    }
    //Test for lane 2 mask checkbox and set/reset lane 2 mask flag
    if ((x>580) && (x<620) && (y>305) && (y<345)) {
      if (ComPort_Avail && PlanCreatedflag) {        //Check if setup complete to run a race
        Mask[1] = !Mask[1];                          //Toggle lane 2 mask state
        if (Mask[0] && Mask[1] && Mask[2]) {         //Are more than 2 lanes masked?
          showMessageDialog(frame,Msg12);            //Display Dialog
          Mask[1] = false;                           //Clear Lane 2 mask flag
        }
        Set_Mask_State();                            //and update the display
        Send_Mask_State();                           //Send mask status to Arduino
      }
      else {
        showMessageDialog(frame,Msg13);              //Display error message dialog
      }
    }
    //Test for lane 3 mask checkbox and set/reset lane 3 mask flag
    if ((x>930) && (x<970) && (y>305) && (y<345)) {
      if (ComPort_Avail && PlanCreatedflag) {        //Check if setup complete to run a race
        Mask[2] = !Mask[2];                            //Toggle lane 3 mask state
        if (Mask[0] && Mask[1] && Mask[2]) {           //Are more than 2 lanes masked?
          showMessageDialog(frame,Msg12);              //Display Dialog
          Mask[2] = false;                             //Clear Lane 3 mask flag
        }
        Set_Mask_State();                              //and update the display
        Send_Mask_State();                             //Send mask status to Arduino
      }
      else {
        showMessageDialog(frame,Msg13);              //Display error message dialog
      }
    }
  }
  if (PageSelect==3) {                               //Do if Results page is selected
    //Test for View Roster button click
    if ((x>50) && (x<180) && (y>150) && (y<220)) {   //View Roster button clicked?
      if (RosterLoadedflag) {                        //Do if Roster is available
        launch(Roster_Filename);                     //Launch Windows command to open the roster
      }
      else {
        prompt = "Roster has not been loaded.\nPlease load a roster first";
        showMessageDialog(frame,prompt);             //Display the prompt for acknowledgement
      }
    }
    //Test for View Race Plan button click
    if ((x>50) && (x<180) && (y>260) && (y<330)) {   //View Race Plan button clicked?
      if (PlanCreatedflag) {                         //Do if race plan is available
        launch(Filename1);                           //Launch Windows command to open the race plan
      }
      else {
        prompt = "Race Plan not available.\nPlease create one first";
        showMessageDialog(frame,prompt);             //Display the prompt for acknowledgement
      }
    }
    //Test for View Race Results button click
    if ((x>50) && (x<180) && (y>370) && (y<440)) {   //View Race Plan button clicked?
      if (RaceResultsAvailflag) {                    //Do if race results are available
        launch(Filename2);                           //Display the prompt for acknowledgement
      }
      else {
        prompt = "Race Results not yet available.\nPlease complete the race first";
        showMessageDialog(frame,prompt);             //Display the prompt for acknowledgement
      }
    }
    //Test for View Overall Race Standings button click
    if ((x>50) && (x<180) && (y>480) && (y<560)) {   //View Overall Standings button clicked?
      if (StandingsAvailflag) {                      //Do if race standings are available
        launch(Filename3);                           //Launch Windows command to open overall standings report
      }
      else {
        prompt = "Race Standings not yet available.\nPlease complete all heats of the race first";
        showMessageDialog(frame,prompt);             //Display the prompt for acknowledgement
      }
    }
    //Test for View Group Race Standings button click
    if ((x>50) && (x<180) && (y>590) && (y<670)) {   //View Group Standings button clicked?
      if (GrpStdngsAvailflag) {                      //Do if Group standings are available
        launch(Filename4);                           //Launch Windows command to open group standings report
      }
      else {
        if (StandingsAvailflag) {
          prompt = "Group Standings were not generated\nbecause there is only 1 group in the race";
        }
        else {
          prompt = "Group Standings not yet available.\nPlease complete all heats of the race first";
        }
        showMessageDialog(frame,prompt);             //Display the prompt for acknowledgement
      }
    }
  }
}

void Draw_SelectCom_Button() {
  fill(200);                                         //Set rectangle color
  rect(50,180,80,60,10);                             //Draw rectangle
  textSize(23);                                      //Set text size to 23 pixels
  textAlign(CENTER);                                 //Set text alignment to center
  fill(0);                                           //Set text color to black
  text("Select",90,205);                             //Display "Select" inside text box top
  text("Com",90,235);                                //Display "Com" inside text box bottom
}

void Draw_LoadRoster_Button() {
  fill(200);                                         //Set rectangle color
  rect(50,280,80,60,10);                             //Draw rectangle
  fill(0);                                           //Set text color to black
  textSize(23);                                      //Set text size to 23 pixels
  if (RosterLoadedflag) {                            //Do if roster has been imported
    textAlign(LEFT);                                 //Set text alignment to center
    text("Roster Imported.  Total Racers: " + Racers, 150, 315);
    fill(150);                                       //Set text color to grey (Disabled) if flag = true
  }
  textAlign(CENTER);                                 //Set text alignment to center
  text("Import",90,305);                             //Display "Import" inside text box top
  text("Roster",90,335);                             //Display "Roster" inside text box bottom
}

void Draw_CreatePlan_Button() {
  fill(200);                                         //Set rectangle color
  rect(50,380,80,60,10);                             //Draw rectangle
  textSize(23);                                      //Set text size to 23 pixels
  textAlign(CENTER);                                 //Set text alignment to center
  fill(0);                                           //Set text color to black
  if (!RosterLoadedflag || PlanCreatedflag) fill(150); //Set text color to grey (Disabled)
  text("Create",90,405);                             //Display "Create" inside text box top
  text("Plan",90,435);                               //Display "Plan" inside text box bottom
  if (PlanCreatedflag) {
    textSize(23);                                    //Set text size to 23 pixels
    textAlign(LEFT);                                 //Set text alignment to center
    fill(0);                                         //Set text color to black
    text("Race Plan Created.   Total Heats in race: " + Total_Heats, 150, 415);
  }
}

void Draw_UpdateTitle_Button() {
  fill(Rx, Gx, Bx);                                  //Set text color background
  noStroke();                                        //Disable drawing of rectangle outline
  rect(145,480,1050,60);                             //Erase any displayed text
  stroke(0);                                         //Enable drawing of rectangle outline
  fill(200);                                         //Set rectangle color
  rect(50,480,80,60,10);                             //Draw rectangle
  textSize(23);                                      //Set text size to 23 pixels
  textAlign(CENTER);                                 //Set text alignment to center
  fill(0);                                           //Set text color to black
  text("Update",90,505);                             //Display "Update" inside text box top
  text("Title",90,535);                              //Display "Title" inside text box bottom
  textAlign(LEFT);                                   //Set text alignment to center
  text("Current Race Page Title: ", 150, 520);       //Draw text
  fill(255, 0, 170);                                 //Set Title color
  text(RP_Title, 380, 520);                          //Draw the current title
}

void Draw_Reset_Button() {
  fill(220);                                         //Set rectangle color
  rect(50,580,80,60,10);                             //Draw rectangle
  textSize(23);                                      //Set text size to 23 pixels
  textAlign(CENTER);                                 //Set text alignment to center
  fill(0);                                           //Set text color to black
  if (!RosterLoadedflag) fill(150);                  //Set text color to grey (Disabled) if no roster isloaded
  text("RESET",90,617);                              //Display "RESET" inside text box
}

void Draw_ViewRoster_Button() {
  fill(200);                                         //Set rectangle color
  rect(50,150,130,70,10);                            //Draw rectangle
  fill(0);                                           //Set text color to black
  if (!RosterLoadedflag) fill(150);                  //Set text color to grey (Disabled) if flag = false
  textSize(23);                                      //Set text size to 23 pixels
  textAlign(CENTER);                                 //Set text alignment to center
  text("View/Print",115,180);                        //Display "Load" inside text box top
  text("Race Roster",115,210);                       //Display "Roster" inside text box bottom
}

void Draw_ViewPlan_Button() {
  fill(200);                                         //Set rectangle color
  rect(50,260,130,70,10);                            //Draw rectangle
  textSize(23);                                      //Set text size to 23 pixels
  textAlign(CENTER);                                 //Set text alignment to center
  fill(0);                                           //Set text color to black
  if (!PlanCreatedflag) fill(150);                   //Set text color to grey (Disabled)
  text("View/Print",115,290);                        //Display "View/Print" inside text box top
  text("Race Plan",115,320);                         //Display "Plan" inside text box bottom
}

void Draw_ViewRace_Results_Button() {
  fill(200);                                         //Set rectangle color
  rect(50,370,130,70,10);                            //Draw rectangle
  textSize(23);                                      //Set text size to 23 pixels
  textAlign(CENTER);                                 //Set text alignment to center
  fill(0);                                           //Set text color to black
  if (!RaceResultsAvailflag) fill(150);              //Set text color to grey (Disabled)
  text("View/Print",115,400);                        //Display "View/Print" inside text box top
  text("Race Results",115,430);                      //Display "Race Results" inside text box bottom
}

void Draw_View_Ovl_Standings_Button() {
  fill(200);                                         //Set rectangle color
  rect(50,480,130,80,10);                            //Draw rectangle
  textSize(23);                                      //Set text size to 23 pixels
  textAlign(CENTER);                                 //Set text alignment to center
  fill(0);                                           //Set text color to black
  if (!StandingsAvailflag) fill(150);                //Set text color to grey (Disabled)
  text("View/Print",115,503);                        //Display "Create" inside text box top
  text("Overall",115,528);                           //Display "Overall" inside text box bottom
  text("Standings",115,553);                         //Display "Standings" inside text box bottom
}

void Draw_View_Grp_Standings_Button() {
  fill(200);                                         //Set rectangle color
  rect(50,590,130,80,10);                            //Draw rectangle
  textSize(23);                                      //Set text size to 23 pixels
  textAlign(CENTER);                                 //Set text alignment to center
  fill(0);                                           //Set text color to black
  if (!GrpStdngsAvailflag) fill(150);                //Set text color to grey (Disabled)
  text("View/Print",115,613);                        //Display "Create" inside text box top
  text("Group",115,638);                             //Display "Group" inside text box bottom
  text("Standings",115,663);                         //Display "Standings" inside text box bottom
}

void Redisplay_RaceResults() {
  //Redisplay time & finish order info for the current selected heat
  Clear_Data();                                      //Call subroutine to clear time & finish order
  if (Heat_Comp[Heat_Number-1]) {                    //Is the selected heat completed?
    //Read times from results table if available
    Index1 = Heat_Number-1;                          //Set the Results table index pointer
    Time[0] = Results.getFloat(Index1, "L1Time");    //Get recorded car time for Lane 1
    Time[1] = Results.getFloat(Index1, "L2Time");    //Get recorded car time for Lane 2
    Time[2] = Results.getFloat(Index1, "L3Time");    //Get recorded car time for Lane 3
    //Determine the lane finish order (place) based on finish times
    for (i=0; i<Num_Lanes; i++) {                    //First set all Place holders to last (3'd) place
      Place[i] = Num_Lanes;
    }
    for (i=0; i<Num_Lanes; i++) {                    //Now compare each time to the other times and
      for (j=0; j<Num_Lanes; j++) {                  //decrement its place value each time it is less
        if (Time[i] < Time[j]) {                     //A value of 9.9999, if present, gets set to last place thus
          Place[i]--;                                //maintaining correct finish order for the valid lanes
        }
      }
    }
    //Write the finish order values into their respective text boxes on the screen if valid
    textAlign(CENTER);                               //Set text alignment to center
    fill(255);                                       //Set text color to white
    textSize(130);                                   //Set text size to 130 pixels
    for (i=0; i<Num_Lanes; i++) {                    //Loop through all lanes
      if (Time[i] != 9.9999) text(Place[i],250+i*350,225);  //Write lane place values
    }
    //Write the time values into their respective text boxes on the screen
    textSize(55);                                    //Set text size to 55 pixels
    for (i=0; i<Num_Lanes; i++) {                    //Loop through all lanes
      Str1=nf(Time[i],1,4);                          //Convert so it won't round to 3 decimal places
      if (Time[i] != 9.9999) text(Str1,250+i*350,290);  //Check for a valid time value and write if so
        else text("- - - - -",255+i*350,285);        //Or write dashes into the lane 1 text box
    }
  }
}

void Display_CarDriver() {
//Display car & driver info for the current heat including lane mask if active
  Clear_Namebox();                                   //Call subroutine to clear names & car numbers
  fill(0);                                           //Set text color to black
  for (i=0; i<Num_Lanes; i++) {                      //Loop through the lanes
    Car[i] = Raceplan.getInt(Heat_Number-1, i*2+1);  //Get car number assigned to each lane
    Driver[i] = Raceplan.getString(Heat_Number-1, i*2+2);  //Get driver name assigned to each lane
    if (Car[i] > 0) {                                //Check for not a BYE lane (Is 0 when a BYE lane)
      Mask[i] = false;
    }
    else {
      Driver[i] = "BYE Lane";                        //Set driver name to BYE Lane
      Mask[i] = true;
    }
    textAlign(CENTER);                               //Set text alignment to center
    textSize(35);                                    //Set text size to 35 pixels
    text(Car[i],145+350*i,387);                      //Write the car number
    textAlign(LEFT);                                 //Set text alignment to left
    textSize(25);                                    //Set text size to 25 pixels
    text(Driver[i],175+350*i,385);                   //Write the driver name
  }
  Set_Mask_State();                                  //Update the display checkbox
  Send_Mask_State();                                 //Send mask status to Arduino
}

void Clear_Namebox() {
  //Draw text boxes with curved corners in which the names and car numbers will be displayed.
  //Note: This will also overwrite (clear) any previous data drawn inside
  fill(200);                                         //Set rectangle color
  for (i=0; i<Num_Lanes; i++) {                      //Loop through all lanes
    rect(155+i*350,355,210,40,10);                   //Draw rectangles for racer name
    ellipse(145+i*350,375,50,50);                    //Draw circle for car number
  }
}

void Draw_Heatbox() {
  //Draw the Heat Count box and current Heat number
  //Note: This will also overwrite (clear) any previous data drawn inside the box
  fill(100);                                         //Set rectangle color
  rect(950,490,160,80);                              //and draw it
  textSize(65);                                      //Set text size to 65 pixels
  textAlign(CENTER);                                 //Set text alignment to center
  fill(255);                                         //Set text color to white
  if (Heat_Number > 0) {
    Str1 = str(Heat_Number) + "/" + str(Total_Heats);//Build the heat number string
  }
  else {
    Str1 = "#/" + str(Total_Heats);                  //Build the heat number string for 0 heats
  }
  text(Str1,1030,555);                               //Display Heat Number & Max Heats
}

void Draw_NextHeat_Button() {
  fill(200);                                         //Set rectangle color
  rect(445,500,80,60,10);                            //Draw rectangle
  textSize(25);                                      //Set text size to 25 pixels
  textAlign(CENTER);                                 //Set text alignment to center
  fill(0);                                           //Set text color to black
  if (!NextHeatflag) fill(150);                      //Set text color to grey (Disabled) if NextHeatflag = FALSE
  if (!ComPort_Avail || !PlanCreatedflag) fill(150); //Set text color to grey if setup incomplete
  text("Next",485,525);                              //Display "Next" inside text box top
  text("Heat",485,555);                              //Display "Heat" inside text box bottom
}

void Draw_RerunHeat_Button() {                       //Subroutine to display the Rerun Heat select button
  fill(200);                                         //Set rectangle color
  rect(565,500,80,60,10);                            //Draw rectangle
  textSize(25);                                      //Set text size to 25 pixels
  textAlign(CENTER);                                 //Set text alignment to center
  fill(0);                                           //Set text color to black
  if (!Rerunflag) fill(150);                         //Set text color to grey (Disabled) if Rerunflag = FALSE
  text("Rerun",605,525);                             //Display "Next" inside text box top
  text("Heat",605,555);                              //Display "Heat" inside text box bottom
}

void Draw_GateRel_Button() {                         //Subroutine to display the Rerun Heat select button
  fill(200);                                         //Set rectangle color
  rect(685,500,80,60,10);                            //Draw rectangle
  textSize(25);                                      //Set text size to 25 pixels
  textAlign(CENTER);                                 //Set text alignment to center
  fill(0);                                           //Set text color to black
  if (!Gaterelflag) fill(150);                       //Set text color to grey (Disabled) if Gaterelflag = FALSE
  text("Gate",725,525);                              //Display "Next" inside text box top
  text("Release",725,555);                           //Display "Heat" inside text box bottom
}

void Set_Mask_State() {
  fill(200);                                         //Set rectangle color
  for (j=0; j<Num_Lanes; j++) {                      //Loop through all lanes
    rect(230+j*350,305,40,40,10);                    //Draw rectangle for Lane mask select boxes
  }
  stroke(255, 0, 0);                                 //Set line color to red
  strokeWeight(5);                                   //Set line weight (thickness)
  for (j=0; j<Num_Lanes; j++) {                      //Loop through all lanes
    if (Mask[j]) {                                   //Check if lane mask is set
      line(237+j*350,325,250+j*350,342);             //Draw check-mark leg 1 if so
      line(250+j*350,342,263+j*350,310);             //Draw check-mark leg 2 if so
    }
  }
  stroke(0);                                         //Reset line color to black (default)
  strokeWeight(1);                                   //Restore line weight to default
}

void Send_Mask_State() {
  myPort.write("U");                                 //Send command to unmask all lanes
  myPort.write(lf);                                  //followed by a line-feed
  delay(50);                                         //Wait a while
  for (i=0; i<Num_Lanes; i++) {                      //Loop through all lanes
    if (Mask[i]) {                                   //Check if lane mask is to be set
      myPort.write("M");                             //If so, send mask prefix
      myPort.write(i+49);                            //and send lane number in ASCII
      myPort.write(lf);                              //followed by a line-feed
    }
  }
}

void Set_Race_State() {
  textAlign(LEFT);                                   //Set text alignment to left
  textSize(25);                                      //Set text size to 25 pixels
  fill(200);                                         //Set color to dim circle & text
  ellipse(80,485,25,25);                             //Draw dimmed circle
  text("Track Not Ready",110,495);                   //Draw dimmed text
  ellipse(80,520,25,25);                             //Draw dimmed circle
  text("Ready",110,530);                             //Draw dimmed text
  ellipse(80,555,25,25);                             //Draw dimmed circle
  text("Racing",110,565);                            //Draw dimmed text
  ellipse(80,590,25,25);                             //Draw dimmed circle
  text("Finished",110,600);                          //Draw dimmed text
  switch (Race_State) {                              //Test the race state
    case 0:                                          //Do the following if it is 0
      fill(255,0,0);                                 //Set color to red
      ellipse(80,485,25,25);                         //Draw red circle
      text("Track Not Ready",110,495);               //Draw red text
      break;
    case 1:                                          //Do the following if it is 1
      fill(245, 255, 0);                             //Set color yellow
      ellipse(80,520,25,25);                         //Draw yellow circle
      text("Ready",110,530);                         //Draw yellow text
      break;
    case 2:                                          //Do the following if it is 2
      fill(0, 180, 0);                               //Set color to dark green
      ellipse(80,555,25,25);                         //Draw green circle
      text("Racing",110,565);                        //Draw green text
      break;
    case 3:                                          //Do the following if it is 3
      fill(0, 0, 200);                               //Set color to blue
      ellipse(80,590,25,25);                         //Draw blue circle
      text("Finished",110,600);                      //Draw blue text
      break;
    default:                                         //Do nothing if it is something else
      //do nothing
      break;
  }
}

void Disp_Message(int Msg_Number) {
  Clear_Msgbox();                                    //Call Clear Message Box subroutine
  textAlign(LEFT);                                   //Set text alignment to left
  textSize(25);                                      //Set text size to 25 pixels
  fill(255);                                         //Set color to white
  switch (Msg_Number) {                              //Get the message number
    case 0:                                          //Do nothing if it is 0
      //do nothing
      break;
    case 1:                                          //Do the following if 1
      text(Msg1,50,677);                             //Display message
      break;
    case 2:                                          //Do the following if 2
      text(Msg2,50,677);                             //Display message
      break;
    case 3:                                          //Do the following if 3
      text(Msg3,50,677);                             //Display message
      break;
    case 4:                                          //Do the following if 4
      text(Msg4,50,677);                             //Display message
      break;
    case 5:                                          //Do the following if 5 (Not yet implemented)
      text(Msg5,50,677);                             //Display message
      break;
    case 6:                                          //Do the following if 6
      text(Msg6,50,677);                             //Display message
      break;
    case 7:                                          //Do the following if 7
      text(Msg7,50,677);                             //Display message
      break;
    default:                                         //Do nothing if something else
      //do nothing
      break;
  }
}

void Clear_Data() {
  //Draw text boxes with curved corners in which the lane times and finish order will be displayed.
  //Note: This will also overwrite (clear) any previous data drawn inside
  fill(100);                                         //Set rectangle color
  for (i=0; i<Num_Lanes; i++) {                      //Loop through all lanes
    rect(210+i*350,120,80,115,10);                   //Draw rectangles for lane finish order
    rect(160+i*350,245,180,50,10);                   //Draw rectangles for lane time
  }
}

void Clear_Msgbox() {
  //Draw rectangle message box in which the error messages will be displayed
  //Note: This will also overwrite (clear) any previous data drawn inside
  fill(100);                                         //Set rectangle color
  rect(40,650,950,40,10);                            //Draw it
}

void Load_Roster() {
  selectInput("Select the Roster file to use:", "RosterSelect"); //Open window to select roster file
  while (!RosterFile_Selected) {
    delay(1000); 
  }
  //NOTE: The following routine gets the roster file location so it can save the race docs in the same folder
  for (i=Roster_Filename.length()-1; i>0; i--) {     //Loop thru roster filename from end to beginning
    Str1 = Roster_Filename.charAt(i)+" ";            //Get character & add a space to make it a string
    if (Str1.equals("\\ ")) {                        //Check it for the backslash "\" character
      File_Path = Roster_Filename.substring(0, i+1); //Extract file path from roster filename
      Filename1 = File_Path + "Race_Plan_"+Date_String+".doc";
      Filename2 = File_Path + "Race_Results_"+Date_String+".csv";
      Filename3 = File_Path + "Race_Standings_Overall_"+Date_String+".csv";
      Filename4 = File_Path + "Race_Standings_Group_"+Date_String+".csv";
      break;                                         //Exit the loop
    }
  }
  //println(File_Path);                              //For debug only
  Roster = loadTable(Roster_Filename, "header");     //Load roster table from file
  Racers = Roster.getRowCount();                     //Get the number of racers in the roster
  if (Racers<2) {                                    //Check for valid roster file
    prompt = "Invalid Roster File - Must have at least 2 racers.\nFix & restart program.";
    showMessageDialog(frame,prompt);                 //Display the prompt for acknowledgement
    exit();                                          //And exit the program
  }
  if (Racers>MaxRacers) {                            //Check for valid roster file
    prompt = "Invalid Roster File - Can't have more than 40 racers.\nFix & restart program.";
    showMessageDialog(frame,prompt);                 //Display the prompt for acknowledgement
    exit();                                          //And exit the program
  }
  Errorflag1 = false;                                //Preset error flag1 to false
  Errorflag2 = false;                                //Preset error flag2 to false
  for (i=0; i<Racers; i++) {                         //Loop through all racers
    kar[i] = Roster.getInt(i,"CarNo");               //Get racer car number
    if (kar[i]<1 || kar[i]>99) Errorflag1 = true;    //And check if valid range & set error flag if not
  }
  for (i=0; i<Racers; i++) {                         //Loop through all racers again
    for (j=i+1; j<Racers; j++) {                     //2nd Loop through the racers
      if (kar[i] == kar[j]) Errorflag2 = true;       //Set error flag if two car numbers are the same
    }
  }
  if (Errorflag1) {                                  //Did error flag 1 get set?
    prompt = "- Invalid Car Number Detected -\nCar numbers must be between 1 and 99.\nFix Roster & restart program.";
    showMessageDialog(frame,prompt);                 //Display the prompt for acknowledgement
    exit();                                          //And exit the program
  }
  if (Errorflag2) {                                  //Did error flag 2 get set?
    prompt = "- Repeat Car Numbers Detected -\nNo two Car numbers can be the same.\nFix Roster & restart program.";
    showMessageDialog(frame,prompt);                 //Display the prompt for acknowledgement
    exit();                                          //And exit the program
  }
  if (!Errorflag1 && !Errorflag2) {                  //Do if no errors found
    //Initialize the Standings table with names, car numbers and groups from the roster table
    Index1 = -1;                                     //Reset the Table row index
    for (i=0; i<Racers; i++) {                       //Loop through all racers in the Roster
      Str1 = Roster.getString(i, "CarNo");           //Get Car Number
      Str2 = Roster.getString(i, "LastName");        //Get racer last name
      Str3 = Roster.getString(i, "FirstName");       //Get racer last name
      Name = Str3 + " " + Str2;                      //Combine first & last name
      Str3 = Roster.getString(i, "Group");           //Get racer Group assignment
      Standings.addRow();                            //Add new row to the Standings table
      Index1++;                                      //Increment the table row index
      Standings.setString(Index1,"CarNo",Str1);      //Write car number to Standings table
      Standings.setString(Index1,"Driver",Name);     //Write racer name to Standings table
      Standings.setString(Index1,"Group",Str3);      //Write racer Group assignment to Standings table
    }
    //Preset lane masking variables
    for (i=0; i<Num_Lanes; i++) {                    //Loop through all lanes
      Mask[i] = false;                               //and set its Mask flag to false
    }
    //Initialize the Heat Complete status array
    for (i=0; i<Max_Heats; i++) {                    //Loop through entire heat complete status array
      Heat_Comp[i] = false;                          //and set its flag to false
    }
  }
}

void Create_RacePlan() {
  //PART 1: Generate the race plan table
  Index1 = 0;                                        //Preset the PPN table row index pointer
  Index2 = 0;                                        //Preset the Raceplan table row index pointer
  while (Index1 < PPN_Table.getRowCount()) {         //Loop through the PPN_Table rows
    if (Racers == (PPN_Table.getInt(Index1, "Cars"))) {  //And search for a match
      HeatNum = PPN_Table.getInt(Index1, "Heat");    //Get heat number
      R1 = PPN_Table.getInt(Index1, "Lane1");        //Get Racer number for Lane 1
      R2 = PPN_Table.getInt(Index1, "Lane2");        //Get Racer number for Lane 2
      R3 = PPN_Table.getInt(Index1, "Lane3");        //Get Racer number for Lane 3
      if (R1>0) {                                    //Do if lane 1 is not a BYE lane
        L1 = Roster.getInt(R1-1,"CarNo");            //Get racer car number to be assigned to Lane 1
        Str1 = Roster.getString(R1-1,"FirstName");   //Get racer car first name to be assigned to Lane 1
        Str2 = Roster.getString(R1-1,"LastName");    //Get racer car last name to be assigned to Lane 1
        Str3 = Str1 + " " + Str2;                    //Combine racer's first & last name
        if (Str3.length()>19) N1 = Str3.substring(0,19); //Trim to 19 characters if too long
          else N1 = Str3;
      }
      else {                                         //Do if lane 1 is a BYE lane
        L1=0;                                        //Set Lane 1 car number to 0 if a BYE lane
        N1 = "BYE";                                  //Set Lane 1 name to "BYE" if a BYE lane
      }
      if (R2>0) {                                    //Do if lane 2 is not a BYE lane
        L2 = Roster.getInt(R2-1,"CarNo");            //Get racer car number to be assigned to Lane 2
        Str1 = Roster.getString(R2-1,"FirstName");   //Get racer car first name to be assigned to Lane 2
        Str2 = Roster.getString(R2-1,"LastName");    //Get racer car last name to be assigned to Lane 2
        Str3 = Str1 + " " + Str2;                    //Combine racer's first & last name
        if (Str3.length()>19) N2 = Str3.substring(0,19); //Trim to 19 characters if too long
          else N2 = Str3;
      }
      else {                                         //Do if lane 2 is a BYE lane
        L2=0;                                        //Set Lane 2 car number to 0 if a BYE lane
        N2 = "BYE";                                  //Set Lane 2 name to "BYE" if a BYE lane
      }
      if (R3>0) {                                    //Do if lane 3 is not a BYE lane
        L3 = Roster.getInt(R3-1,"CarNo");            //Get racer car number to be assigned to Lane 3
        Str1 = Roster.getString(R3-1,"FirstName");   //Get racer car first name to be assigned to Lane 3
        Str2 = Roster.getString(R3-1,"LastName");    //Get racer car last name to be assigned to Lane 3
        Str3 = Str1 + " " + Str2;                    //Combine racer's first & last name
        if (Str3.length()>19) N3 = Str3.substring(0,19); //Trim to 19 characters if too long
          else N3 = Str3;
      }
      else {                                         //Do if lane 3 is a BYE lane
        L3=0;                                        //Set Lane 3 car number to 0 if a BYE lane
        N3 = "BYE";                                  //Set Lane 3 name to "BYE" if a BYE lane
      }
      //Populate the race plan & results table
      Raceplan.addRow();                             //Add a row to the Raceplan table
      Raceplan.setInt(Index2, "Heat", HeatNum);      //Save Heat Number
      Raceplan.setInt(Index2, "L1Car", L1);          //Save Car Number assigned to Lane 1 in race plan table
      Raceplan.setString(Index2, "L1Name", N1);      //Save Name assigned to Lane 1 in race plan table
      Raceplan.setInt(Index2, "L2Car", L2);          //Save Car Number assigned to Lane 2 in race plan table
      Raceplan.setString(Index2, "L2Name", N2);      //Save Name assigned to Lane 2 in race plan table
      Raceplan.setInt(Index2, "L3Car", L3);          //Save Car Number assigned to Lane 3 in race plan table
      Raceplan.setString(Index2, "L3Name", N3);      //Save Name assigned to Lane 3 in race plan table
      Results.addRow();                              //Add a row to the Results table
      Results.setInt(Index2, "Heat", HeatNum);       //Save Heat Number to results table
      Results.setInt(Index2, "L1Car", L1);           //Save Car Number assigned to Lane 1 in results table
      Results.setInt(Index2, "L2Car", L2);           //Save Car Number assigned to Lane 2 in results table
      Results.setInt(Index2, "L3Car", L3);           //Save Car Number assigned to Lane 3 in results table
      Index2++;                                      //Increment index pointer
    }
    Index1++;
  }
  Total_Heats = Raceplan.getRowCount();              //Get the Total_Heats value
  
  //PART 2: Generate the race plan text file
  RP_output = createWriter(Filename1);               //Open new text file in which to save Race Plan
  RP_Header = RP_Title + " Race Plan  "+nf(Mo,2)+"/"+nf(Day,2)+"/"+nf(Yr,4)+"  "+nf(Hr,2)+":"+nf(Min,2);
  RP_output.println(RP_Header);                      //Write header line to the file
  for (i=0; i<Total_Heats; i++) {                    //Cycle through all heats
    RP_output.println("");                           //Insert a blank line
    Str1 = "Heat: "+nf(i+1,2);                       //Build the heat number string
    RP_output.println(Str1);                         //Print the Heat number
    RP_output.println(RP_Text1);                     //Print Header text line 1
    RP_output.println(RP_Text2);                     //Print Header text line 2
    L1 = Raceplan.getInt(i,"L1Car");                 //Get car# for lane 1
    N1 = Raceplan.getString(i,"L1Name");             //Get name for lane 1
    Str1 = "|  1 | "+PadNameField(N1)+"|  "+nf(L1,2)+"  |"; //Build line of text
    RP_output.println(Str1);                         //Print lane 1 racer info
    L2 = Raceplan.getInt(i,"L2Car");                 //Get car# for lane 2
    N2 = Raceplan.getString(i,"L2Name");             //Get name for lane 2
    Str1 = "|  2 | "+PadNameField(N2)+"|  "+nf(L2,2)+"  |"; //Build line of text
    RP_output.println(Str1);                         //Print lane 2 racer info
    L3 = Raceplan.getInt(i,"L3Car");                 //Get car# for lane 3
    N3 = Raceplan.getString(i,"L3Name");             //Get name for lane 3
    Str1 = "|  3 | "+PadNameField(N3)+"|  "+nf(L3,2)+"  |"; //Build line of text
    RP_output.println(Str1);                         //Print lane 3 racer info
    if (i>0 && i+1<Total_Heats && (i+1) % 7 == 0) {  //Have 7 heats been printed?
      RP_output.println(char(12));                   //Insert page break
      RP_output.println(RP_Header);                  //Write header for the next page
    }
  }
  RP_output.flush();                                 //Write the remaining data to the file
  RP_output.close();                                 //Finish & close the file
}

//Subroutine: 
String PadNameField(String Nm) {                     //Enter with racer name
  j = Nm.length();                                   //Get length
  if (j<19) Nm = Nm + Spaces.substring(0, 19-j);     //Add space string to end of name if <19
  return Nm;                                         //And return to calling routine
}

void RosterSelect(File selection) {                  //Used in roster file selection popup
  if (selection == null) RosterFile_Selected = false;
  else {
    Roster_Filename = selection.getAbsolutePath();
    RosterFile_Selected = true;
  }
}

void Populate_Standings_Table() {
  //Pull race results from Results table and populate the Standings table
  for (i=0; i<Standings.getRowCount(); i++) { //Loop through table to clear out any previously saved data
    Standings.setFloat(i,"TotalTime",0.0);           //Clear TotalTime value
    Standings.setInt(i,"HeatsRan",0);                //Clear HeatsRan value
    Standings.setFloat(i,"AvgTime",0.0);             //Clear AvgTime value
    Standings.setFloat(i,"FastTime",9.9999);         //Preset FastTime value to 9.9999
    Standings.setInt(i,"IntTime",0);                 //Clear Integer TotalTime value
  }
  for (i=0; i<Results.getRowCount(); i++) {          //Loop through all Results table rows (heats)
    for(j=0; j<Num_Lanes; j++) {                     //Loop through all cars in each row
      n = Results.getInt(i,j*2+1);                   //Get Car number
      Time1 = Results.getFloat(i,j*2+2);             //Get Car time from Results table
      Itime1 = int(Time1 * 10000);                   //Convert to integer for total time calculations
      if ((Time1>0.0) && (Time1<=9.9999)) {          //Do only if car has a valid time
        for (k=0; k<Standings.getRowCount(); k++) {  //Loop through all cars in the Standings table
          if (n == Standings.getInt(k, "CarNo")) {   //Find the car in the Standings table
            Itime2 = Standings.getInt(k, "IntTime"); //Get previously saved integer total time
            Itime2 = Itime2 + Itime1;                //Add next found time to it
            Standings.setInt(k, "IntTime",Itime2);   //And re-save it
            Time2 = Standings.getFloat(k, "FastTime");  //Get previously saved fastest time
            if (Time1 < Time2) {                     //Check if new time is faster
              Standings.setFloat(k, "FastTime",Time1);  //And update if so
            }
            h = Standings.getInt(k, "HeatsRan");     //Get the current heats ran count
            Standings.setInt(k, "HeatsRan", h+1);    //and bump it by 1
            Time2 = (Itime2/(h+1))/10000.0;          //Calculate new average integer time
            Standings.setFloat(k, "AvgTime",Time2);  //And save it
          }
        }
      }
    }   
  }
  Standings.sort("IntTime");                         //Sort table by overall standings (i.e. Total Time)
  for (k=0; k<Standings.getRowCount(); k++) {        //Loop through all cars in the Standings table
    Standings.setInt(k, "Position", k+1);            //Set the position number
    Itime2 = Standings.getInt(k, "IntTime");         //Get the total time in integer format
    Time1 = Itime2/10000.0;                          //Convert back to floating point
    Standings.setFloat(k, "TotalTime",Time1);        //And re-save it as a floating point value
  }
  Standings.removeColumn("IntTime");                 //Remove the IntTime column before saving
  saveTable(Standings, Filename3);                   //Save the sorted Standings table
  StandingsAvailflag = true;                         //Set standings available flag
  
  //Check table for more than one race group and build Group Standings table if so
  GrpStdngsAvailflag = false;                        //Preset group flag to false
  for (k=0; k<Standings.getRowCount()-1; k++) {      //Loop through all entries in the Standings table
    if (Standings.getString(k, "Group").equals(Standings.getString(k+1, "Group"))) {
      //Do nothing if equal
    }
    else GrpStdngsAvailflag = true;                  //Set group flag to true if more than one group
  }
  if (GrpStdngsAvailflag) {  //If more than one group exists reformat table into Group Standings table
    Standings.sort("Group");                         //Sort table by race group
    Sortflag = true;                                 //Preset flag to ensure at least once through the loop
    while (Sortflag) {                               //Do 'Bubble Sort' while sort flag is true
      Sortflag = false;                              //Set sort flag to false
      for (k=0; k<Standings.getRowCount()-1; k++) {  //Loop through all entries in the Standings table
        if (Standings.getString(k, "Group").equals(Standings.getString(k+1, "Group"))) {
          Time1 = Standings.getFloat(k, "TotalTime");
          if (Time1 > Standings.getFloat(k+1, "TotalTime")) {  //Compare times and swap row data if greater
            x = Standings.getInt(k, "CarNo");
            Name = Standings.getString(k, "Driver");
            Time2 = Standings.getFloat(k, "AvgTime");
            Time3 = Standings.getFloat(k, "FastTime");
            j = Standings.getInt(k, "HeatsRan");
            Standings.setInt(k, "CarNo",Standings.getInt(k+1, "CarNo"));
            Standings.setString(k, "Driver",Standings.getString(k+1, "Driver"));
            Standings.setFloat(k, "TotalTime",Standings.getFloat(k+1, "TotalTime"));
            Standings.setFloat(k, "AvgTime",Standings.getFloat(k+1, "AvgTime"));
            Standings.setFloat(k, "FastTime",Standings.getFloat(k+1, "FastTime"));
            Standings.setInt(k, "HeatsRan",Standings.getInt(k+1, "HeatsRan"));
            Standings.setInt(k+1, "CarNo", x);
            Standings.setString(k+1, "Driver", Name);
            Standings.setFloat(k+1, "TotalTime", Time1);
            Standings.setFloat(k+1, "AvgTime", Time2);
            Standings.setFloat(k+1, "FastTime", Time3);
            Standings.setInt(k+1, "HeatsRan", j);
            Sortflag = true;                         //Set so we go through the loop again
          }
        }
      }
    }
    //Add Position number to each car in the Group Standings table
    i=1;                                             //Initialize Position number
    Standings.setInt(0, "Position", i);              //Set the position number for the first row
    for (k=1; k<Standings.getRowCount(); k++) {      //Loop through 2'nd thru last row in the table
      if (Standings.getString(k, "Group").equals(Standings.getString(k-1, "Group"))) {  //Same group as last row?
        i++;                                         //Increment the Position count
      }
      else i=1;                                      //Reset the position number if not the same group
      Standings.setInt(k, "Position", i);            //Set the position number
    }
    saveTable(Standings, Filename4);                 //Save the Group Standings table
  }
}

void Select_Comm() {
  ComPort_Avail = false;                             //Default COM port available to false
    //Check & Open the serial port at 9600 baud and buffer data till linefeed received
  try {                                              //Use the 'try' & 'catch' exception handling process
    i = Serial.list().length;                        //Get the number of open ports
    if (i != 0) {                                    //Do if at least one port is open
      if (i >= 2) {                                  //Do if more than one open port
        //Create a list of found COM ports from which the user can pick
        for (j = 0; j < i; j++) {
          COMlist += char(j+'a') + " = " + Serial.list()[j];
          if (j+1 < i) COMlist += ",  ";
        }
        COMx = showInputDialog("Which COM port is correct? (a,b,..):\n"+COMlist);
        if (COMx == null) ComPort_Avail = false;;
        if (COMx.isEmpty()) ComPort_Avail = false;;
        i = int(COMx.toLowerCase().charAt(0) - 'a') + 1;
      }
      Portname = Serial.list()[i-1];                 //Get the selected port name
      myPort = new Serial(this, Portname, 9600);     //Activate COM port at 9600 baud
      myPort.bufferUntil('\n');                      //Buffer until Line Feed appears
      prompt = "COM Port Selected: " + Portname;     //Construct COM message & COM Port
      showMessageDialog(frame,prompt);               //Display the prompt for acknowledgement
      ComPort_Avail = true;                          //Set Com Port avail flag
    }
    else {
      showMessageDialog(frame,"No open/connected COM ports detected.\nFix problem & try again");
    }
  }
  catch (Exception e) { //Error catch routine. Print the type of error
    showMessageDialog(frame,"COM port is not available\n(may be in use by another program)");
    //println("Error:", e);                          //For debug only
    //exit();                                        //For debug only
  }
}
